
//  Copyright <2021> <Hans de Nivelle>
//
//  This software is released with the 3-clause BSD license, which 
//  is given below: 
// 
//  Redistribution and use in source and binary forms, with or without 
//  modification, are permitted provided that the following conditions are met:

//  Redistributions of source code must retain the above copyright notice, 
//  this list of conditions and the following disclaimer:

//  Redistributions in binary form must reproduce the above copyright notice, 
//  this list of conditions and the following disclaimer in the documentation 
//  and/or other materials provided with the distribution.
//  Neither the name of the copyright holder nor the names of its contributors 
//  may be used to endorse or promote products derived from this software 
//  without specific prior written permission.

//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
//  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
//  HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
//  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
//  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
//  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
//  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.



// Written by Hans de Nivelle, September 2021.

#include "functional.h" 

bool prolog::functional::equal_to( const term& t ) const
{
   auto p = t. try_cast<functional> ( );
   if(p)
   {
      if( f != p -> f || sub. size( ) != p -> sub. size( ))
         return false; 

      for( size_t i = 0; i != sub. size( ); ++ i )
      {
         if( ! sub[i] -> equal_to( p -> sub[i] ))
            return false;
      }
      return true; 
   }
   return false;
}

void prolog::functional::print( std::ostream& out ) const 
{
   out << f. getstring( );
   if( sub. size( ))
   {
      out << "( ";
      for( auto p = sub. begin( ); p != sub. end( ); ++ p )
      { 
         if( p != sub. begin( ))
            out << ", ";
         (*p) -> print( out );
      }
      out << " )"; 
   }
}

void prolog::functional::printhash( util::hashbuilder& hh ) const 
{
   function::hash h; 

   hh << h(f); 
   for( const auto& s : sub )
      s -> printhash(hh);
}



