
//  Copyright <2021> <Hans de Nivelle>
//
//  This software is released with the 3-clause BSD license, which 
//  is given below: 
// 
//  Redistribution and use in source and binary forms, with or without 
//  modification, are permitted provided that the following conditions are met:

//  Redistributions of source code must retain the above copyright notice, 
//  this list of conditions and the following disclaimer:

//  Redistributions in binary form must reproduce the above copyright notice, 
//  this list of conditions and the following disclaimer in the documentation 
//  and/or other materials provided with the distribution.
//  Neither the name of the copyright holder nor the names of its contributors 
//  may be used to endorse or promote products derived from this software 
//  without specific prior written permission.

//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
//  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
//  HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
//  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
//  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
//  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
//  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.



#ifndef PROLOG_FUNCTIONAL_
#define PROLOG_FUNCTIONAL_  1

#include <iostream> 
#include <vector>

#include "term.h"

#include "function.h"

namespace prolog 
{

   class functional : public termbase
   {

      function f;
      std::vector< term > sub;

   public: 
      functional( function f, std::vector< term > && sub = { } )
         : f(f),
           sub( std::move( sub ))
      { }

      functional( function f, const std::vector< term > & sub )
         : f(f),
           sub( sub )
      { }

      functional( const std::string& f, std::vector< term > && sub = { } )
         : f( function( f, sub. size( ))),
           sub( std::move( sub ))
      { }

      const function& getfunction( ) const { return f; }
      size_t getarity( ) const { return sub. size( ); }

      const term& operator[] ( size_t i ) const { return sub[i]; }

      bool equal_to( const term& t ) const override;
      void print( std::ostream& out ) const override;
      void printhash( util::hashbuilder& hh ) const override;

   };

}

#endif


