
//  Copyright <2021> <Hans de Nivelle>
//
//  This software is released with the 3-clause BSD license, which 
//  is given below: 
// 
//  Redistribution and use in source and binary forms, with or without 
//  modification, are permitted provided that the following conditions are met:

//  Redistributions of source code must retain the above copyright notice, 
//  this list of conditions and the following disclaimer:

//  Redistributions in binary form must reproduce the above copyright notice, 
//  this list of conditions and the following disclaimer in the documentation 
//  and/or other materials provided with the distribution.
//  Neither the name of the copyright holder nor the names of its contributors 
//  may be used to endorse or promote products derived from this software 
//  without specific prior written permission.

//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
//  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
//  HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
//  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
//  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
//  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
//  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.



#include "function.h"
#include "term.h"
#include "opdef.h"
#include "syntax.h"
#include "listconstr.h"
#include "tokenizer.h"
#include "parser.h"

void tptp( prolog::syntax& synt )
{
   synt. insert( { "$", prolog::fx, 99 } );
   synt. insert( { "!", prolog::fx, 400 } ); 
   synt. insert( { "?", prolog::fx, 400 } ); 

   synt. insert( { "=", prolog::xfx, 405 } ); 
   synt. insert( { "!=", prolog::xfx, 405 } ); 

   synt. insert( { "~", prolog::fy, 450 } ); 
   synt. insert( { ":", prolog::xfy, 450 } ); 

   synt. insert( { "|", prolog::xfy, 502 } );
   synt. insert( { "~|", prolog::xfy, 502 } ); 

   synt. insert( { "&", prolog::xfy, 503 } );
   synt. insert( { "~&", prolog::xfy, 503 } ); 

   synt. insert( { "=>", prolog::xfy, 504 } );
   synt. insert( { "<=", prolog::xfy, 504 } );

   synt. insert( { "<=>", prolog::xfy, 505  } );
   synt. insert( { "<~>", prolog::xfy, 505  } );
}


// Taken from Sterling and Shapiro, but corrected, because
// it is very incomplete and contains many errors.

void standard( prolog::syntax& synt )
{
   synt. insert( prolog::opdef( ":-", prolog::xfx, 1200 ));
   synt. insert( prolog::opdef( "-->", prolog::xfx, 1200 ));
   synt. insert( prolog::opdef( ":-", prolog::xf, 1200 ));
      // Sterling & Shapiro say this is a prefix operator, but
      // that cannot be right, because you want to be able to write
      // p(X) :- . 

   synt. insert( prolog::opdef( ";", prolog::xfy, 1100 ));

   synt. insert( { ",", prolog::xfy, 1000 } );

   synt. insert( { "\\+", prolog::fx, 800 } );

   synt. insert( { "=", prolog::xfx, 700 } );
   synt. insert( { "\\=", prolog::xfx, 700 } );

   synt. insert( { "==", prolog::xfx, 700 } );
   synt. insert( { "\\==", prolog::xfx, 700 } );

   synt. insert( { "=..", prolog::xfx, 700 } );

   synt. insert( { "is", prolog::xfx, 700 } );
      // Not present in Sterling & Shapiro.

   synt. insert( { "=:=", prolog::xfx, 700 } );
   synt. insert( { "=\\=", prolog::xfx, 700 } );
   
   synt. insert( { "<", prolog::xfx, 700 } );
   synt. insert( { "=<", prolog::xfx, 700 } );
   synt. insert( { ">", prolog::xfx, 700 } );
   synt. insert( { ">=", prolog::xfx, 700 } );

   synt. insert( prolog::opdef( "+", prolog::yfx, 500 ));
   synt. insert( prolog::opdef( "-", prolog::yfx, 500 ));

   synt. insert( prolog::opdef( "*", prolog::yfx, 400 ));
   synt. insert( prolog::opdef( "/", prolog::yfx, 400 ));
   synt. insert( prolog::opdef( "mod", prolog::yfx, 400 ));

   synt. insert( prolog::opdef( "&", prolog::xfy, 200 ));
   synt. insert( prolog::opdef( "^", prolog::xfy, 200 ));

   synt. insert( prolog::opdef( "-", prolog::fy, 200 ));
}

void addoperator( prolog::syntax& synt, const prolog::listconstr& consnil, 
                  const bigint& attr, const std::string& assoc, 
                  const prolog::term& op )
{

   if( attr > bigint( 100000000 ))
      return;

   unsigned int u = attr. getunsigned( );

   const static std::map< std::string, prolog::optag > 
      assocnames = { { "fx", prolog::fx },
                     { "fy", prolog::fy },
                     { "xfx", prolog::xfx },
                     { "xfy", prolog::xfy },
                     { "yfx", prolog::yfx },
                     { "xf", prolog::xf },
                     { "yf", prolog::yf } };

   auto p = assocnames. find( assoc );
   if( p == assocnames. end( ))
      return;

   auto f = op. try_cast< prolog::functional > ( );
   if( f -> getfunction( ). getarity( ) == 0 )
   {
      const auto& s = f -> getfunction( ). getstring( );

      synt. insert( prolog::opdef( s, p -> second, u )); 
      return;
   }

   if( f -> getfunction( ) == consnil. cons )
   {
      while( f -> getfunction( ) == consnil. cons )
      {
         if( auto car = (*f)[0]. try_cast< prolog::functional > ( ); 
             car && car -> getarity( ) == 0 ) 
         {
            const auto& arg = car -> getfunction( ). getstring( );    

            synt. insert( prolog::opdef( arg, p -> second, u )); 
         }

         f = (*f)[1]. try_cast< prolog::functional > ( ); 
      }

      if( f -> getfunction( ) != consnil. nil )
         std::cout << "it should have been " << consnil. nil << "\n"; 

      return;
   } 
}


int main( int argc, char* argv[] )
{
   filereader rd( &std::cin, "std::cin" ); 

   prolog::tokenizer tok( std::move( rd ));

   prolog::syntax synt;

   std::cout << synt << "\n";

   prolog::listconstr consnil( { "nil", 0 }, { "cons", 2 } );

   std::cout << consnil << "\n";

   varstore< prolog::term > vs;

   prolog::parser prs( tok, vs, synt, consnil );

   prs. debug = 1;
   prs. checkattrtypes = 2;
   prs. checkmovable = 2;
 
   unsigned int nrerrors = 0;

   while( true )
   {
      std::cout << "--------------------------------------\n";
      std::cout << synt << "\n";

      prs. ensurelookahead( );
      if( prs. getlookahead( ). type == prolog::sym_EOF )
      {
         std::cout << "end-of-file reached in normal fashion\n";
         std::cout << "there were " << nrerrors << " errors\n";
         return 0;
      }

      auto res = prs. parse( prolog::sym_Start );
      std::cout << "parser returned:    " << res << "\n\n";
      std::cout << "lookahead = " << prs. getlookahead( ) << "\n"; 

      if( res. type == prolog::sym_OneTerm && 
          prs. getlookahead( ). type == prolog::sym_TERMINATOR )
      {
         std::cout << "this looks like a good parse\n"; 

         // check for 'halt' and 'op': 

         if( auto p = res. get< prolog::term > ( ). 
                      try_cast< prolog::functional > ( ) ; p ) 
         {
            if( p -> getarity( ) == 0 && 
                p -> getfunction( ). getstring( ) == "halt" ) 
            {
               std::cout << "this was the halt command\n";
               std::cout << "there were " << nrerrors << " errors\n";
               return 0;
            }

            if( p -> getarity( ) == 3 &&
                p -> getfunction( ). getstring( ) == "op" )
            {
               auto prior = (*p)[0].try_cast< prolog::constant< bigint >> ( );
               auto assoc = (*p)[1].try_cast< prolog::functional > ( );
               
               if( prior && assoc && assoc -> getarity( ) == 0 )
               {
                  addoperator( synt, consnil, prior -> c, 
                               assoc -> getfunction( ). getstring( ), 
                               (*p)[2] ); 
               }
            }

            if( p -> getarity( ) == 0 &&
                p -> getfunction( ). getstring( ) == "show" )
            {
               std::cout << synt << "\n";
            }

         }
      }
      else
      {
         std::cout << "this looks error number " << ( ++ nrerrors ) << "\n";
         while( prs. getlookahead( ). type != prolog::sym_TERMINATOR &&
                prs. getlookahead( ). type != prolog::sym_EOF )
         {
            tok. contextstack. clear( ); 
               // Becaue the tokenizer remembers context.

            std::cout << "ignoring " << prs. getlookahead( ) << "\n";

            prs. resetlookahead( );
            prs. ensurelookahead( ); 
         }

      }

      if( prs. getlookahead( ). type == prolog::sym_TERMINATOR )
         prs. resetlookahead( ); 
      prs. reset( ); 
   }
 
   return 0; 
}


