
//  Copyright <2021> <Hans de Nivelle>
//
//  This software is released with the 3-clause BSD license, which 
//  is given below: 
// 
//  Redistribution and use in source and binary forms, with or without 
//  modification, are permitted provided that the following conditions are met:

//  Redistributions of source code must retain the above copyright notice, 
//  this list of conditions and the following disclaimer:

//  Redistributions in binary form must reproduce the above copyright notice, 
//  this list of conditions and the following disclaimer in the documentation 
//  and/or other materials provided with the distribution.
//  Neither the name of the copyright holder nor the names of its contributors 
//  may be used to endorse or promote products derived from this software 
//  without specific prior written permission.

//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
//  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
//  HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
//  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
//  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
//  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
//  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.



// The code below was generated by Maphoon 2022.

// Definition of struct symbol:

#ifndef PROLOG_SYMBOL_
#define PROLOG_SYMBOL_    1

#include <iostream>
#include <variant>
#include <optional>
#include <stdexcept>

#line 32 "grammar.m"
 #include "term.h" 
 #include "syntax.h" 
 #include "listconstr.h" 

namespace prolog { 

   enum symboltype
   {
      sym__recover_, sym_COMMA, sym_Prefix, sym_SomeTerms, 
      sym_INTEGER, sym_DOUBLE, sym_TERMINATOR, sym_VARIABLE, 
      sym_RSQBRACKET, sym_BAR, sym_LSQBRACKET, sym_Postfix, 
      sym_COMMENT, sym_Infix, sym_Term, sym_OneTerm, 
      sym_QUOTEDIDENTIFIER, sym_ListEnd, sym_MaybeTerms, sym_Start, 
      sym_GLUEDQUOTEDIDENTIFIER, sym_ERROR, sym_GLUEDIDENTIFIER, sym_WHITESPACE, 
      sym_LPAR, sym_RPAR, sym_EOF, sym_IDENTIFIER
   };

   const char* getcstring( symboltype );

   inline std::ostream& operator << ( std::ostream& out, symboltype tp )
      { out << getcstring( tp );  return out; }

   struct symbol
   {
      using attrtype = std::variant < std::monostate, term, double, 
            std::vector<term>, std::string, bigint, opdef > ;

      symboltype type;
      attrtype attr;

      symbol( ) = delete;
      symbol( const symbol& ) = default;
      symbol( symbol&& ) noexcept = default;
      symbol& operator = ( const symbol& ) = default;
      symbol& operator = ( symbol&& ) noexcept = default;
      ~symbol( ) = default;

      symbol( symboltype type )
       : type( type )
      { }

      symbol( symboltype type,
              attrtype && attr )
       : type( type ),
         attr( std::move( attr ))
      { }

      symbol( symboltype type,
              const term & attr )
       : type( type ),
         attr( attr )
      { }

      symbol( symboltype type,
              term && attr )
       : type( type ),
         attr( std::move( attr ))
      { }

      symbol( symboltype type,
              double attr )
       : type( type ),
         attr( attr )
      { }

      symbol( symboltype type,
              const std::vector<term> & attr )
       : type( type ),
         attr( attr )
      { }

      symbol( symboltype type,
              std::vector<term> && attr )
       : type( type ),
         attr( std::move( attr ))
      { }

      symbol( symboltype type,
              const std::string & attr )
       : type( type ),
         attr( attr )
      { }

      symbol( symboltype type,
              std::string && attr )
       : type( type ),
         attr( std::move( attr ))
      { }

      symbol( symboltype type,
              const bigint & attr )
       : type( type ),
         attr( attr )
      { }

      symbol( symboltype type,
              bigint && attr )
       : type( type ),
         attr( std::move( attr ))
      { }

      symbol( symboltype type,
              const opdef & attr )
       : type( type ),
         attr( attr )
      { }

      symbol( symboltype type,
              opdef && attr )
       : type( type ),
         attr( std::move( attr ))
      { }


      void print( std::ostream& out ) const;

      bool has_right_attribute( ) const;

      template< typename T > T& get( )
         { return std::get<T> ( attr ); }
      
      template< typename T > const T& get( ) const &
         { return std::get<T> ( attr ); }
      
      template< typename T > const T& get( ) &&
         { return std::get<T> ( std::move( attr ) ); }
      
   };

   inline
   std::ostream& operator << ( std::ostream& out, const symbol& sym )
      { sym. print( out ); return out; }


}

#endif



