
//  Copyright <2021> <Hans de Nivelle>
//
//  This software is released with the 3-clause BSD license, which 
//  is given below: 
// 
//  Redistribution and use in source and binary forms, with or without 
//  modification, are permitted provided that the following conditions are met:

//  Redistributions of source code must retain the above copyright notice, 
//  this list of conditions and the following disclaimer:

//  Redistributions in binary form must reproduce the above copyright notice, 
//  this list of conditions and the following disclaimer in the documentation 
//  and/or other materials provided with the distribution.
//  Neither the name of the copyright holder nor the names of its contributors 
//  may be used to endorse or promote products derived from this software 
//  without specific prior written permission.

//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
//  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
//  HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
//  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
//  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
//  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
//  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.



// Written by Hans de Nivelle, Sept 2021.

#ifndef PROLOG_TERM_
#define PROLOG_TERM_  1

#include "util/shared_ptr.h"
#include "util/hashbuilder.h"

#include "bigint.h"

#include <unordered_set>

namespace prolog
{
   struct termbase;

   using term = util::shared_ptr< termbase > ; 

   struct termbase
   {
      termbase( ) = default; 
      termbase( termbase&& ) noexcept = default; 
      termbase( const termbase& ) = default;
      termbase& operator = ( termbase&& ) noexcept = default; 
      termbase& operator = ( const termbase& ) = default;
      virtual ~termbase( ) = default; 

      virtual bool equal_to( const term& ) const = 0; 
      virtual void print( std::ostream& out ) const = 0; 
      virtual void printhash( util::hashbuilder& hh ) const = 0;
   };

   inline std::ostream& operator << ( std::ostream& out, const term& t )
   {
      t -> print( out );
      return out;
   }

   inline 
   util::hashbuilder& operator << ( util::hashbuilder& hh, const term& t )
   {
      t -> printhash( hh );
      return hh;
   }

   struct term_hash
   {
      size_t operator( ) ( const term& t ) const
         { util::hashbuilder b; t -> printhash(b); return b.val; }
   };

   struct term_equal_to
   {
      bool operator( ) ( const term& t1, const term& t2 ) const
         { return t1 -> equal_to( t2 ); } 
   };

   using setofadjectives = 
      std::unordered_set< term, term_hash, term_equal_to > ;

}

#include "variable.h" 
#include "constant.h"
#include "functional.h"

#endif 



