
//  Copyright <2021> <Hans de Nivelle>
//
//  This software is released with the 3-clause BSD license, which 
//  is given below: 
// 
//  Redistribution and use in source and binary forms, with or without 
//  modification, are permitted provided that the following conditions are met:

//  Redistributions of source code must retain the above copyright notice, 
//  this list of conditions and the following disclaimer:

//  Redistributions in binary form must reproduce the above copyright notice, 
//  this list of conditions and the following disclaimer in the documentation 
//  and/or other materials provided with the distribution.
//  Neither the name of the copyright holder nor the names of its contributors 
//  may be used to endorse or promote products derived from this software 
//  without specific prior written permission.

//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
//  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
//  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
//  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
//  HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
//  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
//  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
//  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
//  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.



#ifndef PROLOG_TOKENIZER_
#define PROLOG_TOKENIZER_    1

#include "filereader.h"


#include "symbol.h"

namespace prolog
{

   struct tokenizer
   {
      filereader read;

      enum context { ct_list, ct_func, ct_par, ct_glued };
         // ct_list : inside [ ... ]
         // ct_func:  inside f( ... )
         // ct_par:   inside ( ... )
         // ct_glued:  after f in f( ... ).

      static const char* getcstring( context c );
  
      std::vector< context > contextstack; 
         // This is used for characters like | . and , 

      
      tokenizer( ) 
         : read( filereader( ))
      { }

      tokenizer( filereader&& read )
         : read( std::move( read ))
      { }

      symbol get( const syntax& synt );

      void print( std::ostream& out ) const; 

      void test( const syntax& synt );
         // Keep on reading and printing symbols until we reach EOF. 
   }; 

   inline 
   std::ostream& operator << ( std::ostream& out, const tokenizer& tok )
   {
      tok. print( out );
      return out;
   }

}

#endif


